from __future__ import annotations
from typing import Union, NamedTuple, Any, List
from functools import wraps
from supervisely.app.widgets import Widget
from supervisely.app.content import StateJson, DataJson

"""
size1 = 10
x1 = list(range(size1))
y1 = np.random.randint(low=10, high=148, size=size1).tolist()
s1 = [{"x": x, "y": y} for x, y in zip(x1, y1)]

size2 = 30
x2 = list(range(size2))
y2 = np.random.randint(low=0, high=300, size=size2).tolist()
s2 = [{"x": x, "y": y} for x, y in zip(x2, y2)]

chart = sly.app.widgets.Apexchart(
    series=[{"name": "Max", "data": s1}, {"name": "Denis", "data": s2}],
    options={
        "chart": {"type": "line", "zoom": {"enabled": False}},
        "dataLabels": {"enabled": False},
        # "stroke": {"curve": "straight"},
        "stroke": {"curve": "smooth", "width": 2},
        "title": {"text": "Product Trends by Month", "align": "left"},
        "grid": {"row": {"colors": ["#f3f3f3", "transparent"], "opacity": 0.5}},
        "xaxis": {"type": "category"},
    },
    type="line",
)
"""


class Apexchart(Widget):
    class Routes:
        CLICK = "chart_clicked_cb"

    class ClickedDataPoint(NamedTuple):
        series_index: int
        series_name: str
        data_index: int
        data: dict
        x: Any
        y: Any

    def __init__(
        self,
        series: list,
        options: dict,
        type: str,
        height: Union[int, str] = "300",
        sly_options={},
    ):
        self._series = series
        self._options = options
        self._type = type
        self._height = height
        self._click_handled = False
        self._sly_options = sly_options
        super().__init__(file_path=__file__)

    def get_json_data(self):
        return {
            "series": self._series,
            "options": self._options,
            "type": self._type,
            "height": self._height,
            "sly_options": self._sly_options,
        }

    def get_json_state(self):
        return {"clicked_value": None}

    def get_clicked_value(self):
        return StateJson()[self.widget_id]["clicked_value"]

    def get_clicked_datapoint(self) -> Apexchart.ClickedDataPoint:
        value = self.get_clicked_value()
        series_index = value["seriesIndex"]
        data_index = value["dataPointIndex"]
        if series_index == -1 and data_index != -1:
            # zero point (0,0) click
            series_index = 0
        if series_index == -1 or data_index == -1:
            return
        series_name = self._series[series_index]["name"]
        data = self._series[series_index]["data"][data_index]
        res = Apexchart.ClickedDataPoint(
            series_index, series_name, data_index, data, data["x"], data["y"]
        )
        return res

    def click(self, func):
        route_path = self.get_route_path(Apexchart.Routes.CLICK)
        server = self._sly_app.get_server()

        self._click_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_clicked_datapoint()
            func(res)

        return _click

    def add_series(self, name: str, x: list, y: list, send_changes=True):
        if len(x) != len(y):
            raise ValueError(
                f"Lists x and y have different lenght, {len(x)} != {len(y)}"
            )
        data = [{"x": px, "y": py} for px, py in zip(x, y)]
        series = {"name": name, "data": data}
        self._series.append(series)
        self.update_data()
        if send_changes:
            DataJson().send_changes()



# coding: utf-8
"""create/download/update :class:`Dataset<supervisely.project.project.Dataset>`"""

# docs
from __future__ import annotations
from typing import List, NamedTuple, Optional, Dict

import urllib
from supervisely.api.module_api import (
    ApiField,
    ModuleApi,
    UpdateableModule,
    RemoveableModuleApi,
)


class DatasetInfo(NamedTuple):
    """ """

    id: int
    name: str
    description: str
    size: int
    project_id: int
    images_count: int
    items_count: int
    created_at: str
    updated_at: str
    reference_image_url: str


class DatasetApi(UpdateableModule, RemoveableModuleApi):
    """
    API for working with :class:`Dataset<supervisely.project.project.Dataset>`. :class:`DatasetApi<DatasetApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        project_id = 1951
        ds = api.dataset.get_list(project_id)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple DatasetInfo information about Dataset.

        :Example:

         .. code-block:: python

            DatasetInfo(id=452984,
                        name='ds0',
                        description='',
                        size='3997776',
                        project_id=118909,
                        images_count=11,
                        items_count=11,
                        created_at='2021-03-03T15:54:08.802Z',
                        updated_at='2021-03-16T09:31:37.063Z',
                        reference_image_url='https://app.supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/images/original/K/q/jf/...png')
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.SIZE,
            ApiField.PROJECT_ID,
            ApiField.IMAGES_COUNT,
            ApiField.ITEMS_COUNT,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.REFERENCE_IMAGE_URL,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **DatasetInfo**.
        """
        return "DatasetInfo"

    def __init__(self, api):
        ModuleApi.__init__(self, api)
        UpdateableModule.__init__(self, api)

    def get_list(
        self, project_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[DatasetInfo]:
        """
        List of Datasets in the given Project.

        :param project_id: Project ID in which the Datasets are located.
        :type project_id: int
        :param filters: List of params to sort output Datasets.
        :type filters: List[dict], optional
        :return: List of all Datasets with information for the given Project. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[DatasetInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 1951

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()
            ds = api.dataset.get_list(project_id)

            print(ds)
            # Output: [
            #     DatasetInfo(id=2532,
            #                 name="lemons",
            #                 description="",
            #                 size="861069",
            #                 project_id=1951,
            #                 images_count=6,
            #                 items_count=6,
            #                 created_at="2021-03-02T10:04:33.973Z",
            #                 updated_at="2021-03-10T09:31:50.341Z",
            #                 reference_image_url="http://app.supervise.ly/z6ut6j8bnaz1vj8aebbgs4-public/images/original/...jpg"),
            #                 DatasetInfo(id=2557,
            #                 name="kiwi",
            #                 description="",
            #                 size="861069",
            #                 project_id=1951,
            #                 images_count=6,
            #                 items_count=6,
            #                 created_at="2021-03-10T09:31:33.701Z",
            #                 updated_at="2021-03-10T09:31:44.196Z",
            #                 reference_image_url="http://app.supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/images/original/...jpg")
            # ]
        """
        return self.get_list_all_pages(
            "datasets.list",
            {ApiField.PROJECT_ID: project_id, ApiField.FILTER: filters or []},
        )

    def get_info_by_id(self, id: int) -> DatasetInfo:
        """
        Get Datasets information by ID.

        :param id: Dataset ID in Supervisely.
        :type id: int
        :return: Information about Dataset. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`DatasetInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            dataset_id = 384126

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            ds_info = api.dataset.get_info_by_id(dataset_id)
        """
        return self._get_info_by_id(id, "datasets.info")

    def create(
        self,
        project_id: int,
        name: str,
        description: Optional[str] = "",
        change_name_if_conflict: Optional[bool] = False,
    ) -> DatasetInfo:
        """
        Create Dataset with given name in the given Project.

        :param project_id: Project ID in Supervisely where Dataset will be created.
        :type project_id: int
        :param name: Dataset Name.
        :type name: str
        :param description: Dataset description.
        :type description: str, optional
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :return: Information about Dataset. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`DatasetInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 116482

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            ds_info = api.dataset.get_list(project_id)
            print(len(ds_info)) # 1

            new_ds = api.dataset.create(project_id, 'new_ds')
            new_ds_info = api.dataset.get_list(project_id)
            print(len(new_ds_info)) # 2
        """
        effective_name = self._get_effective_new_name(
            parent_id=project_id,
            name=name,
            change_name_if_conflict=change_name_if_conflict,
        )
        response = self._api.post(
            "datasets.add",
            {
                ApiField.PROJECT_ID: project_id,
                ApiField.NAME: effective_name,
                ApiField.DESCRIPTION: description,
            },
        )
        return self._convert_json_info(response.json())

    def get_or_create(
        self, project_id: int, name: str, description: Optional[str] = ""
    ) -> DatasetInfo:
        """
        Checks if Dataset with given name already exists in the Project, if not creates Dataset with the given name.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param name: Dataset name.
        :type name: str
        :param description: Dataset description.
        :type description: str, optional
        :return: Information about Dataset. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`DatasetInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 116482

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            ds_info = api.dataset.get_list(project_id)
            print(len(ds_info)) # 1

            api.dataset.get_or_create(project_id, 'ds1')
            ds_info = api.dataset.get_list(project_id)
            print(len(ds_info)) # 1

            api.dataset.get_or_create(project_id, 'new_ds')
            ds_info = api.dataset.get_list(project_id)
            print(len(ds_info)) # 2
        """
        dataset_info = self.get_info_by_name(project_id, name)
        if dataset_info is None:
            dataset_info = self.create(project_id, name, description=description)
        return dataset_info

    def _get_update_method(self):
        """ """
        return "datasets.editInfo"

    def _remove_api_method_name(self):
        """ """
        return "datasets.remove"

    def copy_batch(
        self,
        dst_project_id: int,
        ids: List[int],
        new_names: Optional[List[str]] = None,
        change_name_if_conflict: Optional[bool] = False,
        with_annotations: Optional[bool] = False,
    ) -> List[DatasetInfo]:
        """
        Copy given Datasets to the destination Project by IDs.

        :param dst_project_id: Destination Project ID in Supervisely.
        :type dst_project_id: int
        :param ids: IDs of copied Datasets.
        :type ids: List[int]
        :param new_names: New Datasets names.
        :type new_names: List[str], optional
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :param with_annotations: If True copies Datasets with annotations, otherwise copies just items from Datasets without annotations.
        :type with_annotations: bool, optional
        :raises: :class:`RuntimeError` if can not match "ids" and "new_names" lists, len(ids) != len(new_names)
        :return: Information about Datasets. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[DatasetInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dst_proj_id = 1980
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 0

            ds_ids = [2532, 2557]
            ds_names = ["lemon_test", "kiwi_test"]

            copied_datasets = api.dataset.copy_batch(dst_proj_id, ids=ds_ids, new_names=ds_names, with_annotations=True)
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 2
        """
        if new_names is not None and len(ids) != len(new_names):
            raise RuntimeError(
                'Can not match "ids" and "new_names" lists, len(ids) != len(new_names)'
            )

        new_datasets = []
        for idx, dataset_id in enumerate(ids):
            dataset = self.get_info_by_id(dataset_id)
            new_dataset_name = dataset.name
            if new_names is not None:
                new_dataset_name = new_names[idx]
            src_images = self._api.image.get_list(dataset.id)
            src_image_ids = [image.id for image in src_images]
            new_dataset = self._api.dataset.create(
                dst_project_id,
                new_dataset_name,
                dataset.description,
                change_name_if_conflict=change_name_if_conflict,
            )
            self._api.image.copy_batch(
                new_dataset.id, src_image_ids, change_name_if_conflict, with_annotations
            )
            new_datasets.append(new_dataset)
        return new_datasets

    def copy(
        self,
        dst_project_id: int,
        id: int,
        new_name: Optional[str] = None,
        change_name_if_conflict: Optional[bool] = False,
        with_annotations: Optional[bool] = False,
    ) -> DatasetInfo:
        """
        Copies given Dataset in destination Project by ID.

        :param dst_project_id: Destination Project ID in Supervisely.
        :type dst_project_id: int
        :param id: ID of copied Dataset.
        :type id: int
        :param new_name: New Dataset name.
        :type new_name: str, optional
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :param with_annotations: If True copies Dataset with annotations, otherwise copies just items from Dataset without annotation.
        :type with_annotations: bool, optional
        :return: Information about Dataset. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`DatasetInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dst_proj_id = 1982
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 0

            new_ds = api.dataset.copy(dst_proj_id, id=2540, new_name="banana", with_annotations=True)
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 1
        """
        new_datasets = self.copy_batch(
            dst_project_id, [id], [new_name], change_name_if_conflict, with_annotations
        )
        if len(new_datasets) == 0:
            return None
        return new_datasets[0]

    def move_batch(
        self,
        dst_project_id: int,
        ids: List[int],
        new_names: Optional[List[str]] = None,
        change_name_if_conflict: Optional[bool] = False,
        with_annotations: Optional[bool] = False,
    ) -> List[DatasetInfo]:
        """
        Moves given Datasets to the destination Project by IDs.

        :param dst_project_id: Destination Project ID in Supervisely.
        :type dst_project_id: int
        :param ids: IDs of moved Datasets.
        :type ids: List[int]
        :param new_names: New Datasets names.
        :type new_names: List[str], optional
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :param with_annotations: If True moves Datasets with annotations, otherwise moves just items from Datasets without annotations.
        :type with_annotations: bool, optional
        :raises: :class:`RuntimeError` if can not match "ids" and "new_names" lists, len(ids) != len(new_names)
        :return: Information about Datasets. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[DatasetInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dst_proj_id = 1978
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 0

            ds_ids = [2545, 2560]
            ds_names = ["banana_test", "mango_test"]

            movied_datasets = api.dataset.move_batch(dst_proj_id, ids=ds_ids, new_names=ds_names, with_annotations=True)
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 2
        """
        new_datasets = self.copy_batch(
            dst_project_id, ids, new_names, change_name_if_conflict, with_annotations
        )
        self.remove_batch(ids)
        return new_datasets

    def move(
        self,
        dst_project_id: int,
        id: int,
        new_name: Optional[str] = None,
        change_name_if_conflict: Optional[bool] = False,
        with_annotations: Optional[bool] = False,
    ) -> DatasetInfo:
        """
        Moves given Dataset in destination Project by ID.

        :param dst_project_id: Destination Project ID in Supervisely.
        :type dst_project_id: int
        :param id: ID of moved Dataset.
        :type id: int
        :param new_name: New Dataset name.
        :type new_name: str, optional
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :param with_annotations: If True moves Dataset with annotations, otherwise moves just items from Dataset without annotation.
        :type with_annotations: bool, optional
        :return: Information about Dataset. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`DatasetInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dst_proj_id = 1985
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 0

            new_ds = api.dataset.move(dst_proj_id, id=2550, new_name="cucumber", with_annotations=True)
            ds = api.dataset.get_list(dst_proj_id)
            print(len(ds)) # 1
        """
        new_dataset = self.copy(
            dst_project_id, id, new_name, change_name_if_conflict, with_annotations
        )
        self.remove(id)
        return new_dataset

    def _convert_json_info(self, info: dict, skip_missing=True):
        """ """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        if res.reference_image_url is not None:
            res = res._replace(reference_image_url=res.reference_image_url)
        if res.items_count is None:
            res = res._replace(items_count=res.images_count)
        return DatasetInfo(**res._asdict())



import functools
import os
import pathlib
import shutil
import tempfile
import threading
import traceback
from collections import namedtuple
from distutils.dir_util import copy_tree

from fastapi import FastAPI
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles

import supervisely as sly


def get_static_paths_by_mounted_object(mount) -> list:
    StaticPath = namedtuple("StaticPath", ["local_path", "url_path"])
    static_paths = []

    if hasattr(mount, "routes"):
        for current_route in mount.routes:
            if type(current_route) == Mount and type(current_route.app) == FastAPI:
                all_children_paths = get_static_paths_by_mounted_object(current_route)
                for index, current_path in enumerate(all_children_paths):
                    current_url_path = pathlib.Path(
                        str(current_route.path).lstrip("/"),
                        str(current_path.url_path).lstrip("/"),
                    )
                    all_children_paths[index] = StaticPath(
                        local_path=current_path.local_path, url_path=current_url_path
                    )
                static_paths.extend(all_children_paths)
            elif (
                type(current_route) == Mount and type(current_route.app) == StaticFiles
            ):
                static_paths.append(
                    StaticPath(
                        local_path=pathlib.Path(current_route.app.directory),
                        url_path=pathlib.Path(str(current_route.path).lstrip("/")),
                    )
                )

    return static_paths


def dump_statics_to_dir(static_dir_path: pathlib.Path, static_paths: list):
    for current_path in static_paths:
        current_local_path: pathlib.Path = current_path.local_path
        current_url_path: pathlib.Path = static_dir_path / current_path.url_path

        if current_local_path.is_dir():
            current_url_path.mkdir(parents=True, exist_ok=True)
            copy_tree(
                current_local_path.as_posix(),
                current_url_path.as_posix(),
                preserve_symlinks=True,
            )


def dump_html_to_dir(static_dir_path, template):
    pathlib.Path(static_dir_path / template.template.name).write_bytes(template.body)


def upload_to_supervisely(static_dir_path):
    api: sly.Api = sly.Api.from_env()

    team_id, task_id = int(os.getenv("context.teamId")), int(os.getenv("TASK_ID", 0000))
    remote_dir = pathlib.Path("/", "offline-sessions", str(task_id), "app-template")

    res_remote_dir: str = api.file.upload_directory(
        team_id=team_id,
        local_dir=static_dir_path.as_posix(),
        remote_dir=remote_dir.as_posix(),
    )

    if os.getenv("TASK_ID") is not None:
        api.task.update_meta(id=task_id, data={"templateRootDirectory": res_remote_dir})

    sly.logger.info(f"App files stored in {res_remote_dir} for offline usage")


def dump_files_to_supervisely(app: FastAPI, template_response):
    try:
        if (
            os.getenv("_SUPERVISELY_OFFLINE_FILES_UPLOADED", "False") == "True"
            and template_response.context.get("request") is not None
        ):
            return
        os.environ["_SUPERVISELY_OFFLINE_FILES_UPLOADED"] = "True"
        sly.logger.info(f"Saving app files for offline usage")

        app_template_path = pathlib.Path(tempfile.mkdtemp())
        app_static_paths = get_static_paths_by_mounted_object(mount=app)
        dump_statics_to_dir(
            static_dir_path=app_template_path, static_paths=app_static_paths
        )
        dump_html_to_dir(static_dir_path=app_template_path, template=template_response)

        upload_to_supervisely(static_dir_path=app_template_path)

        shutil.rmtree(app_template_path.as_posix())

    except Exception as ex:
        traceback.print_exc()
        sly.logger.warning(f"Cannot dump files for offline usage, reason: {ex}")
        os.environ["_SUPERVISELY_OFFLINE_FILES_UPLOADED"] = "False"


def available_after_shutdown(app: FastAPI):
    def func_layer_wrapper(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            template_response = f(*args, **kwargs)
            try:
                if sly.utils.is_production():
                    sly.logger.info(f"Start dumping app UI for offline mode")
                    threading.Thread(
                        target=functools.partial(
                            dump_files_to_supervisely, app, template_response
                        ),
                        daemon=False,
                    ).start()

                return template_response
            except Exception as ex:
                traceback.print_exc()
                sly.logger.warning(f"Cannot dump files for offline usage, reason: {ex}")

        return wrapper

    return func_layer_wrapper



# coding: utf-8
"""list/create supervisely workspaces"""

from __future__ import annotations
from typing import NamedTuple, List, Dict, Optional

from supervisely.api.module_api import ApiField, ModuleApi, UpdateableModule


class WorkspaceInfo(NamedTuple):
    """
    """

    id: int
    name: str
    description: str
    team_id: int
    created_at: str
    updated_at: str


class WorkspaceApi(ModuleApi, UpdateableModule):
    """
    API for working with Workspace. :class:`WorkspaceApi<WorkspaceApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        workspace_info = api.workspace.get_info_by_id(workspace_id) # api usage example
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple WorkspaceInfo containing information about Workspace.

        :Example:

         .. code-block:: python

            WorkspaceInfo(id=15,
                          name='Cars',
                          description='Workspace contains Project with annotated Cars',
                          team_id=8,
                          created_at='2020-04-15T10:50:41.926Z',
                          updated_at='2020-04-15T10:50:41.926Z')
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.TEAM_ID,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **WorkspaceInfo**.
        """
        return 'WorkspaceInfo'

    def __init__(self, api):
        ModuleApi.__init__(self, api)
        UpdateableModule.__init__(self, api)

    def get_list(self, team_id: int, filters: Optional[List[Dict[str, str]]] = None) -> List[WorkspaceInfo]:
        """
        List of Workspaces in the given Team.

        :param team_id: Team ID in which the Workspaces are located.
        :type team_id: int
        :param filters: List of params to sort output Workspaces.
        :type filters: List[dict], optional
        :return: List of all Workspaces with information for the given Team. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[WorkspaceInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            workspace_infos = api.workspace.get_list(8)
            print(workspace_infos)
            # Output: [
            # WorkspaceInfo(id=15,
            #               name='Cars',
            #               description='',
            #               team_id=8,
            #               created_at='2020-04-15T10:50:41.926Z',
            #               updated_at='2020-04-15T10:50:41.926Z'),
            # WorkspaceInfo(id=18,
            #               name='Heart',
            #               description='',
            #               team_id=8,
            #               created_at='2020-05-20T15:01:54.172Z',
            #               updated_at='2020-05-20T15:01:54.172Z'),
            # WorkspaceInfo(id=20,
            #               name='PCD',
            #               description='',
            #               team_id=8,
            #               created_at='2020-06-24T11:51:11.336Z',
            #               updated_at='2020-06-24T11:51:11.336Z')
            # ]

            # Filtered Workspace list
            workspace_infos = api.workspace.get_list(8, filters=[{ 'field': 'name', 'operator': '=', 'value': 'Heart'}])
            print(workspace_infos)
            # Output: [WorkspaceInfo(id=18,
            #                       name='Heart',
            #                       description='',
            #                       team_id=8,
            #                       created_at='2020-05-20T15:01:54.172Z',
            #                       updated_at='2020-05-20T15:01:54.172Z')
            # ]
        """
        return self.get_list_all_pages('workspaces.list', {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id: int) -> WorkspaceInfo:
        """
        Get Workspace information by ID.

        :param id: Workspace ID in Supervisely.
        :type id: int
        :return: Information about Workspace. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`WorkspaceInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            workspace_info = api.workspace.get_info_by_id(58)
            print(workspace_info)
            # Output: WorkspaceInfo(id=58,
            #                       name='Test',
            #                       description='',
            #                       team_id=8,
            #                       created_at='2020-11-09T18:21:08.202Z',
            #                       updated_at='2020-11-09T18:21:08.202Z')
        """
        return self._get_info_by_id(id, 'workspaces.info')

    def create(self, team_id: int, name: str, description: Optional[str] = "",
               change_name_if_conflict: Optional[bool] = False) -> WorkspaceInfo:
        """
        Create Workspace with given name in the given Team.

        :param team_id: Team ID in Supervisely where Workspace will be created.
        :type team_id: int
        :param name: Workspace Name.
        :type name: str
        :param description: Workspace description.
        :type description: str, optional
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :return: Information about Workspace. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`WorkspaceInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            new_workspace = api.workspace.create(8, "Vehicle Detection")
            print(new_workspace)
            # Output: WorkspaceInfo(id=274,
            #                       name='Vehicle Detection"',
            #                       description='',
            #                       team_id=8,
            #                       created_at='2021-03-11T12:24:21.773Z',
            #                       updated_at='2021-03-11T12:24:21.773Z')
        """
        effective_name = self._get_effective_new_name(
            parent_id=team_id, name=name, change_name_if_conflict=change_name_if_conflict)
        response = self._api.post('workspaces.add', {ApiField.TEAM_ID: team_id,
                                                     ApiField.NAME: effective_name,
                                                     ApiField.DESCRIPTION: description})
        return self._convert_json_info(response.json())

    def _get_update_method(self):
        """
        """
        return 'workspaces.editInfo'

    def _convert_json_info(self, info: dict, skip_missing=True) -> WorkspaceInfo:
        """
        """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return WorkspaceInfo(**res._asdict())



from supervisely.app.widgets import Widget
from supervisely.api.project_api import ProjectInfo
from supervisely.project.project import Project


class ProjectThumbnail(Widget):
    def __init__(self, info: ProjectInfo):
        self._info = info
        self._description = f"{self._info.items_count} {self._info.type}"
        self._url = Project.get_url(self._info.id)
        super().__init__(file_path=__file__)

    def get_json_data(self):
        return {
            "id": self._info.id,
            "name": self._info.name,
            "description": self._description,
            "url": self._url,
            "image_preview_url": self._info.image_preview_url,
        }

    def get_json_state(self):
        return None



# coding: utf-8

import sys
import time
import json
import requests


# should be stateless
class RetrierAbstract:
    def __init__(self, retry_cnt, wait_sec_first, wait_sec_max, timeout, swallow_exc=False):
        self.retry_cnt = int(retry_cnt)
        self.wait_sec = (wait_sec_first, wait_sec_max)
        if isinstance(timeout, list):  # requests lib timeout format
            self.timeout = tuple(timeout)
        else:
            self.timeout = timeout
        self.swallow_exc = swallow_exc

    def _determine_time_to_sleep(self, att_done):
        if att_done > 100:
            res = self.wait_sec[1]
        else:
            res = self.wait_sec[0] * (2 ** (att_done - 1))
            res = min(res, self.wait_sec[1])
        return res

    def _need_raise(self, att_done):
        if att_done < self.retry_cnt:
            time.sleep(self._determine_time_to_sleep(att_done))
        else:
            return not self.swallow_exc

    def request(self, cback, *args, **kwargs):
        raise NotImplementedError()


class RetrierAlways(RetrierAbstract):
    def request(self, cback, *args, **kwargs):
        for att in range(self.retry_cnt):
            try:
                return cback(*args, timeout=self.timeout, **kwargs)
            except Exception:
                if self._need_raise(att + 1):
                    raise
        return None


class RetrierAlwaysYield(RetrierAbstract):
    def request(self, cback, *args, **kwargs):
        for att in range(self.retry_cnt):
            try:
                yield from cback(*args, timeout=self.timeout, **kwargs)
                return
            except Exception:
                if self._need_raise(att + 1):
                    raise
        return None


class RetrierConnTO(RetrierAbstract):
    def request(self, cback, *args, **kwargs):
        for att in range(self.retry_cnt):
            try:
                return cback(*args, timeout=self.timeout, **kwargs)
            except (requests.ConnectionError, requests.ConnectTimeout):
                if self._need_raise(att + 1):
                    raise
        return None


class RetrierConnTOYield(RetrierAbstract):
    def request(self, cback, *args, **kwargs):
        for att in range(self.retry_cnt):
            try:
                yield from cback(*args, timeout=self.timeout, **kwargs)
                return
            except (requests.ConnectionError, requests.ConnectTimeout):
                if self._need_raise(att + 1):
                    raise
        return None


_default_retriers_config = {
    "__endless_stream_in": {
        "class": "RetrierAlwaysYield",
        "params": {
            "retry_cnt": 1000,
            "wait_sec_first": 2,
            "wait_sec_max": 4,
            "timeout": [
                4,
                30
            ]
        }
    },
    "__data_stream_in": {
        "class": "RetrierConnTOYield",
        "params": {
            "retry_cnt": 100,
            "wait_sec_first": 2,
            "wait_sec_max": 4,
            "timeout": [
                4,
                60
            ]
        }
    },
    "__data_stream_out": {
        "class": "RetrierConnTO",
        "params": {
            "retry_cnt": 100,
            "wait_sec_first": 2,
            "wait_sec_max": 4,
            "timeout": [
                4,
                60
            ]
        }
    },
    "__simple_request": {
        "class": "RetrierConnTO",
        "params": {
            "retry_cnt": 1000,
            "wait_sec_first": 1,
            "wait_sec_max": 4,
            "timeout": [
                4,
                60
            ]
        }
    },
    "Log": {
        "class": "RetrierAlways",
        "params": {
            "retry_cnt": 1000,
            "wait_sec_first": 1,
            "wait_sec_max": 1,
            "timeout": [
                4,
                30
            ],
            "swallow_exc": True
        }
    },
    "AgentConnected": {
        "class": "RetrierAlways",
        "params": {
            "retry_cnt": 1000000000,
            "wait_sec_first": 2,
            "wait_sec_max": 20,
            "timeout": [
                4,
                10
            ]
        }
    },
    "AgentPing": {
        "class": "RetrierAlways",
        "params": {
            "retry_cnt": 1000,
            "wait_sec_first": 1,
            "wait_sec_max": 4,
            "timeout": [
                4,
                10
            ]
        }
    }

}


def _retrier_from_cfg(cfg):
    cls_name = cfg['class']
    cls = getattr(sys.modules[__name__], cls_name)
    obj = cls(**cfg['params'])
    return obj


def retriers_from_cfg(cfg_path):
    if not cfg_path:
        cfg = _default_retriers_config
    else:
        cfg = json.load(open(cfg_path, 'r'))
    retriers = {selector: _retrier_from_cfg(item) for selector, item in cfg.items()}  # no validation for now
    return retriers
